from abc import ABC, abstractmethod


class Scorer(ABC):
    """
    Abstract scorer class with an add_to_score method.
    This can be subclassed to implement specific scoring systems.
    """

    def __init__(self):
        self._score = 0
        self._level = 0
        self._lines_cleared = 0

    def add_to_score(self, lines_cleared: int) -> bool:
        """
        Increments the score based on the type of scoring system
        :param lines_cleared: number of lines cleared
        :param level: current level the player is on
        :return: True if reached new level, False otherwise
        """
        assert 1 <= lines_cleared <= 4, f"Lines cleared should be 1-4, can't score {lines_cleared}!"
        return self._add_to_score(lines_cleared)

    @abstractmethod
    def _add_to_score(self, lines_cleared: int) -> bool:
        ...

    @property
    def score(self) -> int:
        return self._score


    @property
    def level(self) -> int:
        return self._level

    @property
    def lines_cleared(self) -> int:
        return self._lines_cleared


class SimpleScorer(Scorer):
    """
    A simple scorer that uses a single rule based on lines cleared,
     without taking level into account.
    """
    _LINES_TO_POINTS = {
        1: 40,
        2: 100,
        3: 300,
        4: 1200,
    }
    _LINES_CLEARED_PER_LEVEL_INCR = 10

    def _add_to_score(self, lines_cleared: int) -> bool:
        self._lines_cleared += lines_cleared
        self._score += self._LINES_TO_POINTS[lines_cleared] * (self.level + 1)
        new_level = self._lines_cleared // self._LINES_CLEARED_PER_LEVEL_INCR
        changed = new_level > self._level
        if changed:
            self._level = new_level
        return changed




import argparse


def parse_args() -> argparse.Namespace:
    """
    Parses the command-line args of the game
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--board_height",
        type=int,
        default=20,
        help="The number of rows that the board should have."
    )
    parser.add_argument(
        "--board_width",
        type=int,
        default=10,
        help="The number of columns that the board should have."
    )
    return parser.parse_args()



from enum import Enum

import pygame


class Command(Enum):
    MOVE_LEFT = 1
    MOVE_RIGHT = 2
    MOVE_DOWN = 3
    MOVE_BOTTOM = 4
    ROTATE = 5
    QUIT = 6
    HELP = 7
    PAUSE = 8

    @classmethod
    def from_char(cls, char: str):
        if not char.isalpha():
            raise ValueError(f"Non-alphabet chars found in input: {char!r}")
        char = char.upper()
        mapping = {
            "L": cls.MOVE_LEFT,
            "R": cls.MOVE_RIGHT,
            "D": cls.MOVE_DOWN,
            "DD": cls.MOVE_BOTTOM,
            "U": cls.ROTATE,
            "Q": cls.QUIT,
            "H": cls.HELP,
            "P": cls.PAUSE,
        }
        if char not in mapping:
            raise ValueError(f"Unsupported input: {char!r}")
        return mapping[char]

    @classmethod
    def from_pygame_key(cls, key: int):
        mapping = {
            pygame.K_LEFT: cls.MOVE_LEFT,
            pygame.K_RIGHT: cls.MOVE_RIGHT,
            pygame.K_DOWN: cls.MOVE_DOWN,
            pygame.K_SPACE: cls.MOVE_BOTTOM,
            pygame.K_UP: cls.ROTATE,
            pygame.K_q: cls.QUIT,
            pygame.K_h: cls.HELP,
            pygame.K_p: cls.PAUSE,
        }
        if key not in mapping:
            raise ValueError(f"Unsupported pygame key: {key}")
        return mapping[key]


from .board import Board
from .engine import EngineCLI
from .scorer import SimpleScorer
from .piece import PieceGenerator
from .utils import parse_args


def main():
    parsed = parse_args()
    board = Board(parsed.board_height, parsed.board_width)
    scorer = SimpleScorer()
    piece_generator = PieceGenerator()
    engine = EngineCLI(board, scorer, piece_generator)
    engine.run()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List

import pygame

from .board import Board
from .colours import BLACK_COLOUR, GREY_COLOUR, RED_COLOUR, WHITE_COLOUR, ORANGE_COLOUR
from .piece import PIECE_COLOURS_RGB, PieceGenerator
from .command import Command
from .scorer import Scorer
from .point import MinoPoint


class Interface(ABC):

    def __init__(self, board: Board, scorer: Scorer, piece_generator: PieceGenerator):
        self._board = board
        self._scorer = scorer
        self._piece_generator = piece_generator

    @abstractmethod
    def draw_screen(self) -> None:
        """
        Draws the main gameplay screen
        :return: None
        """
        ...

    @abstractmethod
    def get_input(self) -> List[Command]:
        """
        Returns any commands the user has inputted
        :return: list of commands
        """
        ...

    @abstractmethod
    def draw_game_over(self) -> None:
        """
        Draws the end of game screen
        :return: None
        """
        ...

    @abstractmethod
    def show_instructions(self) -> None:
        """
        Shows the instructions to the player
        :return: None
        """
        ...

    @abstractmethod
    def quit(self) -> None:
        """
        Ends the session
        :return: None
        """
        ...

    @abstractmethod
    def show_paused(self) -> None:
        """
        Shows the user that the game is paused
        :return: None
        """
        ...


class InterfaceCLI(Interface):
    _INSTRUCTIONS = """
Welcome to the tetris game. You will be prompted for a command at each step.
The possible commands are:
    1. 'L'  -> Move piece left
    2. 'R'  -> Move piece right
    3. 'D'  -> Move piece down
    4. 'DD' -> Move piece as far down as possible
    5. 'U'  -> Rotate piece 90 degrees clockwise
    6. 'Q'  -> Quit
    7. 'H'  -> Bring up this message again
"""

    def draw_screen(self) -> None:
        """
        Draws the main gameplay screen
        :return: None
        """
        print()
        print(f"SCORE: {self._scorer.score}")
        print(f"LINES CLEARED: {self._scorer.lines_cleared}")
        print("Board state:")
        print(self._board)
        print()
        print(f"Next Piece: {self._piece_generator.next_piece_type.letter}")

    def get_input(self) -> List[Command]:
        """
        Returns any commands the user has inputted
        :return: list of commands
        """
        try:
            cmd = Command.from_char(input("Input a command [L/R/D/DD/U/Q/H]: "))
        except ValueError:
            return []
        else:
            return [cmd]

    def draw_game_over(self) -> None:
        """
        Draws the end of game screen
        :return: None
        """
        print("GAME OVER")
        print(f"FINAL SCORE: {self._scorer.score}")
        print(f"LINES CLEARED: {self._scorer.lines_cleared}")
        print(f"LEVEL: {self._scorer.level}")

    def show_instructions(self) -> None:
        """
        Shows the instructions to the player
        :return: None
        """
        print(self._INSTRUCTIONS)

    def quit(self) -> None:
        """
        Ends the session
        :return: None
        """
        print("Quitting...")

    def show_paused(self) -> None:
        """
        Shows the user that the game is paused
        :return: None
        """
        pass


class InterfacePygame(Interface):

    _BLOCK_SCALE_FACTOR = 36
    _WIDTH_PADDING = 40
    _HEIGHT_PADDING = 10
    _INSTRUCTIONS = [
        "← : Move Left",
        "→ : Move Right",
        "↓ : Move Down",
        "↑ : Rotate",
        "<SPC> : Hard Drop",
        "Q : Quit",
        "P : Pause",
    ]
    _font_name = "freesans"
    def __init__(self, *args):
        super().__init__(*args)
        # pygame setup
        pygame.init()
        biggest_screen = sorted(pygame.display.get_desktop_sizes(), reverse=True)[0]
        self._block_size = biggest_screen[1] // self._BLOCK_SCALE_FACTOR

        self._screen_width = (self._board.width + self._WIDTH_PADDING) * self._block_size
        self._screen_height = (self._board.height + self._HEIGHT_PADDING) * self._block_size
        self._screen_size = (self._screen_width, self._screen_height)

        self._grid_width = self._board.width * self._block_size
        self._grid_height = self._board.height * self._block_size
        self._grid_top_left_x = int((self._screen_width - self._grid_width) * 0.50)
        self._grid_top_left_y = (self._screen_height - self._grid_height) // 2

        # Title
        title_label_font_size = int(self._block_size * 1.2)
        title_font = pygame.font.SysFont(self._font_name, title_label_font_size, bold=True)
        self._title_label = title_font.render("TETRIS", 1, WHITE_COLOUR)

        # Game over screen
        self._game_over_label = title_font.render("GAME OVER", 1, RED_COLOUR)

        # Subtitle font, used for score, next piece & pause labels
        subtitle_label_font_size = int(title_label_font_size * 0.6)
        self._subtitle_font = pygame.font.SysFont(self._font_name, subtitle_label_font_size, bold=True)


        # Text Font
        text_font_size = int(title_label_font_size * 0.45)
        self._text_font = pygame.font.SysFont(self._font_name, text_font_size, bold=True)

        # Info Section
        self._info_box_width = self._grid_width
        self._info_box_height = self._grid_height
        self._info_box_top_left_x = int(self._screen_width * 0.75 - self._grid_width // 2)
        self._info_box_top_left_y = self._grid_top_left_y

        self._next_piece_label = self._subtitle_font.render("NEXT PIECE", 1, WHITE_COLOUR)

        self._paused_label = self._subtitle_font.render("PAUSED", 1, ORANGE_COLOUR)
        self._paused_label_top_left_x = self._info_box_top_left_x + self._info_box_width // 2 - self._paused_label.get_width() / 2
        self._paused_label_top_left_y = self._info_box_top_left_y + self._info_box_height - 1.5 * self._block_size

        # Set screen
        self._screen = pygame.display.set_mode(self._screen_size)

    def draw_screen(self) -> None:
        """
        Draws the main gameplay screen
        :return: None
        """
        self._screen.fill(BLACK_COLOUR)
        self._draw_title()
        self._draw_play_grid()
        self._draw_info_section()
        pygame.display.update()

    def _draw_play_grid(self):
        self._draw_tetriminoes()
        self._draw_grid_lines()
        self._draw_border()

    def get_input(self) -> List[Command]:
        """
        Returns any commands the user has inputted
        :return: list of commands
        """
        events = pygame.event.get()
        cmds = []
        for event in events:
            ev_type = event.type
            if ev_type == pygame.QUIT:
                cmds.append(Command.QUIT)
                break
            elif ev_type == pygame.KEYDOWN:
                try:
                    cmd = Command.from_pygame_key(event.key)
                except ValueError:
                    continue
                else:
                    cmds.append(cmd)
                    if cmd is Command.QUIT:
                        break
        return cmds

    def draw_game_over(self) -> None:
        """
        Draws the end of game screen
        :return: None
        """
        self._screen.fill(BLACK_COLOUR)
        self._draw_game_over_title()
        self._draw_final_score()
        pygame.display.update()

    def show_instructions(self) -> None:
        """
        Shows the instructions to the player
        :return: None
        """
        instrs_label = self._subtitle_font.render("INSTRUCTIONS", 1, WHITE_COLOUR)
        top_left_x = self._info_box_top_left_x + self._block_size
        instr_top_left_y = self._info_box_top_left_y + self._info_box_height * 0.55
        self._screen.blit(instrs_label,(self._info_box_top_left_x + self._info_box_width / 2 - instrs_label.get_width() / 2, instr_top_left_y))
        sy = instr_top_left_y + instrs_label.get_height() * 1.5
        for i, l in enumerate(self._INSTRUCTIONS):
            top_left_y = sy + self._text_font.get_height() * i * 1.2
            self._text_font.get_height()
            label = self._text_font.render(l, 1, WHITE_COLOUR)
            self._screen.blit(label, (top_left_x, top_left_y))


    def quit(self) -> None:
        """
        Ends the session
        :return: None
        """
        pygame.display.quit()

    def _draw_title(self) -> None:
        self._screen.blit(
            self._title_label,
            (
                self._screen_width / 2 - (self._title_label.get_width() / 2),
                self._block_size,
            ),
        )

    def _draw_tetriminoes(self):
        sx = self._grid_top_left_x
        sy = self._grid_top_left_y
        for i in range(self._board.height):
            for j in range(self._board.width):
                pygame.draw.rect(
                    surface=self._screen,
                    color=PIECE_COLOURS_RGB[self._board.value_at(i, j)],
                    rect=(sx + j * self._block_size, sy + i * self._block_size, self._block_size, self._block_size),
                    width=0,
                )

    def _draw_grid_lines(self) -> None:
        """
        Draws the grey grid lines that we see
        :return: None
        """
        sx = self._grid_top_left_x
        sy = self._grid_top_left_y
        for i in range(self._board.height):
            pygame.draw.line(
                self._screen,
                GREY_COLOUR,
                (sx, sy + i * self._block_size),
                (sx + self._grid_width - 1, sy + i * self._block_size),
            )  # horizontal lines
            for j in range(self._board.width):
                pygame.draw.line(
                    self._screen,
                    GREY_COLOUR,
                    (sx + j * self._block_size, sy),
                    (sx + j * self._block_size, sy + self._grid_height - 1),
                )  # vertical line

    def _draw_border(self) -> None:
        sx = self._grid_top_left_x
        sy = self._grid_top_left_y
        pygame.draw.rect(surface=self._screen, color=RED_COLOUR, rect=(sx, sy, self._grid_width, self._grid_height), width=2)

    def show_paused(self) -> None:
        """
        Shows the user that the game is paused
        :return: None
        """

        self._screen.blit(self._paused_label, (self._paused_label_top_left_x, self._paused_label_top_left_y))
        pygame.display.update()

    def _draw_info_section(self) -> None:
        # Border
        pygame.draw.rect(
            surface=self._screen,
            color=GREY_COLOUR,
            rect=(self._info_box_top_left_x, self._info_box_top_left_y, self._info_box_width, self._info_box_height),
            width=1
        )
        self._draw_score_stats()
        self._draw_next_piece_section()
        self.show_instructions()

        # Line above pause box
        pause_box_line_separator_y = self._paused_label_top_left_y - (self._info_box_top_left_y + self._info_box_height - (
                    self._paused_label_top_left_y + self._paused_label.get_height()))
        pygame.draw.line(
            self._screen,
            GREY_COLOUR,
            (self._info_box_top_left_x, pause_box_line_separator_y),
            (self._info_box_top_left_x + self._info_box_width, pause_box_line_separator_y)

        )

    def _draw_score_stats(self) -> None:
        score_label = self._subtitle_font.render(
            f"SCORE: {self._scorer.score}",
            1,
            WHITE_COLOUR,
        )
        lines_label = self._subtitle_font.render(
            f"LINES CLEARED: {self._scorer.lines_cleared}",
            1,
            WHITE_COLOUR,
        )
        level_label = self._subtitle_font.render(
            f"LEVEL: {self._scorer.level}",
            1,
            WHITE_COLOUR,
        )
        top_left_x = self._info_box_top_left_x + self._block_size
        score_top_left_y = self._info_box_top_left_y + self._block_size
        lines_top_left_y = score_top_left_y + self._block_size
        level_top_left_y = lines_top_left_y + self._block_size
        self._screen.blit(score_label, (top_left_x, score_top_left_y))
        self._screen.blit(lines_label, (top_left_x, lines_top_left_y))
        self._screen.blit(level_label, (top_left_x, level_top_left_y))

        sec_separator_line_y = level_top_left_y + level_label.get_height() + score_top_left_y - self._info_box_top_left_y
        pygame.draw.line(
            self._screen,
            GREY_COLOUR,
            (self._info_box_top_left_x, sec_separator_line_y),
            (self._info_box_top_left_x + self._info_box_width, sec_separator_line_y),
        )


    def _draw_next_piece_section(self) -> None:
        label_top_left_x = self._info_box_top_left_x + self._info_box_width // 2 - self._next_piece_label.get_width() / 2
        label_top_left_y = self._info_box_top_left_y + self._block_size * 6
        self._screen.blit(self._next_piece_label, (label_top_left_x, label_top_left_y))

        num_blocks_width = 6
        num_blocks_height = 2
        box_width = num_blocks_width * self._block_size
        box_height = num_blocks_height * self._block_size
        box_top_left_x = label_top_left_x + (self._next_piece_label.get_width() - box_width) / 2
        box_top_left_y = label_top_left_y + self._block_size * 1.5

        pygame.draw.rect(
            surface=self._screen,
            color=GREY_COLOUR,
            rect=(box_top_left_x, box_top_left_y, box_width, box_height),
            width=1
        )
        next_piece_type = self._piece_generator.next_piece_type
        blocks, _ = next_piece_type.points_from_top_left(MinoPoint(2, 0))

        for block in blocks:
            pygame.draw.rect(
                surface=self._screen,
                color=PIECE_COLOURS_RGB[next_piece_type.piece_index],
                rect=(box_top_left_x + block.x * self._block_size, box_top_left_y + block.y * self._block_size, self._block_size, self._block_size),
                width=0,
            )

        for i in range(num_blocks_height):
            pygame.draw.line(
                self._screen,
                GREY_COLOUR,
                (box_top_left_x, box_top_left_y + i * self._block_size),
                (box_top_left_x + box_width, box_top_left_y + i * self._block_size),
            )  # horizontal lines
            for j in range(num_blocks_width):
                pygame.draw.line(
                    self._screen,
                    GREY_COLOUR,
                    (box_top_left_x + j * self._block_size, box_top_left_y),
                    (box_top_left_x + j * self._block_size, box_top_left_y + box_height),
                )  # vertical line

        # TODO: next piece section isn't vertically central in its box right now
        separator_line_y = box_top_left_y + (num_blocks_height + 1) * self._block_size
        pygame.draw.line(
            self._screen,
            GREY_COLOUR,
            (self._info_box_top_left_x, separator_line_y),
            (self._info_box_top_left_x + self._info_box_width, separator_line_y),
        )

    def _draw_game_over_title(self) -> None:
        self._screen.blit(
            self._game_over_label,
            (
                self._screen_width / 2 - (self._game_over_label.get_width() / 2),
                self._screen_height * 0.4
            ),
        )

    def _draw_final_score(self) -> None:
        score_label = self._subtitle_font.render(f"FINAL SCORE: {self._scorer.score}", 1, WHITE_COLOUR)
        lines_label = self._subtitle_font.render(f"LINES CLEARED: {self._scorer.lines_cleared}", 1, WHITE_COLOUR)
        level_label = self._subtitle_font.render(f"LEVEL: {self._scorer.level}", 1, WHITE_COLOUR)
        self._screen.blit(
            score_label,
            (
                self._screen_width / 2 - (score_label.get_width() / 2),
                self._screen_height * 0.5
            ),
        )
        self._screen.blit(
            lines_label,
            (
                self._screen_width / 2 - (lines_label.get_width() / 2),
                self._screen_height * 0.5 + score_label.get_height() * 1.5,
            ),
        )
        self._screen.blit(
            level_label,
            (
                self._screen_width / 2 - (level_label.get_width() / 2),
                self._screen_height * 0.5 + (lines_label.get_height() + score_label.get_height()) * 1.5,
            ),
        )



import random
from typing import List, Type

from . import piece
from .point import MinoPoint

Grid = List[List[int]]
_ROW_PADDING = 2


class Board:

    def __init__(self, height: int = 20, width: int = 10):
        # grid with 2 extra rows for generating new pieces at random at the top
        self._height = height + _ROW_PADDING
        self._width = width
        self._grid: Grid = self._new_rows(self._height, width)
        self._next_piece_type = piece.new_piece_type()

    def reached_top_row(self) -> bool:
        """
        Checks if stack has reached the top row
        :return: True if there are blocks in top row, False otherwise
        """
        top_playable_row = self._grid[_ROW_PADDING]
        return sum(top_playable_row) > 0

    def new_piece(self, piece_cls: Type["piece.Piece"]) -> "piece.Piece":
        """
        Adds a new piece of the given type to the board
        :param piece_cls: Piece type
        :return: None
        """
        # TODO: use smarter logic for starting location of piece ?
        # the +2 on the x and -3 on the y coordinates are used here to make sure all pieces can fit
        # e.g 'I' piece has another 3 mino blocks on the right of the top left mino,
        # and L has 2 on the left
        top_left = MinoPoint(random.randint(0 + 2, self._width - 1 - 3), 0)
        piece = piece_cls(self, top_left)
        return piece

    def clear_completed_rows(self, rows: List[int]) -> int:
        """
        Checks if any rows are complete and removes them
        :param rows: the rows to check
        :return: the number of rows cleared
        """
        rows = sorted(rows, reverse=True)
        removed = 0
        for i in rows:
            if self._full_row(i):
                self._remove_row(i)
                removed += 1
        self._grid = self._new_rows(removed, self._width) + self._grid
        return removed

    def update_piece_location(self, piece: "piece.Piece", old_points: List[MinoPoint]) -> None:
        """
        Updates a piece, used in moving pieces
        :param piece: the piece being moved
        :param old_points: the old coordinates of its blocks
        :return: None
        """
        for p in old_points:
            self._set_at_point(p, 0)
        self.add_piece(piece)

    def can_shift(self, piece: "piece.Piece", new_points: List[MinoPoint]) -> bool:
        """
        Checks if a piece can be moved to the new points
        :param piece: the piece
        :param new_points: new points to be moved into
        :return: True if possible to move, False otherwise
        """
        self._remove_piece(piece)
        res = True
        for p in new_points:
            if self._point_outside_grid(p) or self._at_point(p) > 0:
                res = False
                break
        self.add_piece(piece)
        return res

    def add_piece(self, piece: "piece.Piece") -> None:
        """
        Adds a piece to the board
        :param piece: the piece to add
        :return: None
        """
        for p in piece.points:
            assert self._at_point(p) == 0
        for p in piece.points:
            self._set_at_point(p, piece.piece_index)

    def space_below(self, point: MinoPoint) -> bool:
        """
        Checks if there is empty space below a block
        :param point: the coordinate
        :return: True if empty space underneath, False otherwise
        """
        below_y = point.y + 1
        if self._point_outside_grid(MinoPoint(point.x, below_y)):
            return False
        return self._grid[below_y][point.x] == 0

    @property
    def height(self) -> int:
        """
        Public height value used by other components
        :return: integer height
        """
        # subtract the row padding for the public height value used by other components, as this is a detail
        # internal to the board logic. Other components don't need to be aware of it.
        return self._height - _ROW_PADDING

    @property
    def width(self) -> int:
        """
        Public width value used by other components
        :return: integer width
        """
        return self._width

    def value_at(self, i: int, j: int) -> int:
        """
        Gets the value at the coordinate
        :param i: row
        :param j: column
        :return: integer code of the piece at that coordinate
        """
        return self._grid[i + _ROW_PADDING][j]

    def _full_row(self, idx: int) -> bool:
        row = self._grid[idx]
        return all(x > 0 for x in row)

    def _remove_row(self, idx: int) -> None:
        self._grid.pop(idx)

    @staticmethod
    def _new_rows(height: int, width: int) -> Grid:
        return [[0] * width for _ in range(height)]

    def _remove_piece(self, piece: "piece.Piece") -> None:
        for p in piece.points:
            assert self._at_point(p) == piece.piece_index
        for p in piece.points:
            self._set_at_point(p, 0)

    def _at_point(self, p: MinoPoint) -> int:
        return self._grid[p.y][p.x]

    def _set_at_point(self, p: MinoPoint, val: int) -> None:
        self._grid[p.y][p.x] = val

    def _point_outside_grid(self, point: MinoPoint) -> bool:
        return (not (0 <= point.y < self._height)) or (not (0 <= point.x < self._width))

    def __str__(self) -> str:
        res = []
        for row in self._grid[_ROW_PADDING:]:
            row_res = []
            for x in row:
                assert isinstance(x, int)
                marker = "X" if x > 0 else "0"
                row_res.append(piece.PIECE_COLOURS_ANSI[x].format(marker))
            res.append(" ".join(row_res))
        pretty_grid = "\n".join(res)
        return str(pretty_grid)


from enum import Enum

from .command import Command


class Direction(Enum):
    DOWN = 1
    RIGHT = 2
    LEFT = 3

    @classmethod
    def from_command(cls, command: Command):
        mapping = {
            Command.MOVE_LEFT: cls.LEFT,
            Command.MOVE_RIGHT: cls.RIGHT,
            Command.MOVE_DOWN: cls.DOWN,
            Command.MOVE_BOTTOM: cls.DOWN,
        }
        if command not in mapping:
            raise ValueError(f"Command {command} cannot be mapped to a direction")
        return mapping[command]


from dataclasses import dataclass

from .direction import Direction

_DIRECTION_SHIFT = {
    Direction.DOWN: (0, 1),
    Direction.LEFT: (-1, 0),
    Direction.RIGHT: (1, 0),
}
@dataclass
class Point:
    x: float
    y: float

    def shift(self, direction: Direction) -> "Point":
        """
        Creates a new point object with coordinates equivalent to shifting this piece
        :param direction: the direction of movement
        :return: the new point
        """
        cls = type(self)
        x_shift, y_shift = _DIRECTION_SHIFT[direction]
        return cls(self.x + x_shift, self.y + y_shift)


@dataclass
class MinoPoint(Point):
    x: int
    y: int



from abc import ABC, abstractmethod
from typing import Tuple, List, Set, Type
import random

from .direction import Direction

from . import board

from .colours import (
    BLACK_COLOUR,
    CYAN_COLOUR,
    PINK_COLOUR,
    YELLOW_COLOUR,
    GREEN_COLOUR,
    RED_COLOUR,
    BLUE_COLOUR,
    PURPLE_COLOUR,
)
from .point import Point, MinoPoint


PIECE_COLOURS_ANSI = [
    "\033[37m{}\033[00m",  # white, for empty spaces
    "\033[96m{}\033[00m",
    "\033[95m{}\033[00m",
    "\033[93m{}\033[00m",
    "\033[92m{}\033[00m",
    "\033[91m{}\033[00m",
    "\033[34m{}\033[00m",
    "\033[35m{}\033[00m",
]

PIECE_COLOURS_RGB = [
    BLACK_COLOUR,
    CYAN_COLOUR,
    PINK_COLOUR,
    YELLOW_COLOUR,
    GREEN_COLOUR,
    RED_COLOUR,
    BLUE_COLOUR,
    PURPLE_COLOUR
]


class Piece(ABC):
    def __init__(self, board: "board.Board", top_left: MinoPoint):
        ps, c = self.points_from_top_left(top_left)
        self._points: List[MinoPoint] = ps
        self._centre: Point = c
        self._board = board
        self._board.add_piece(self)

    @property
    def points(self) -> List[MinoPoint]:
        """
        :return: List of coordinates for the blocks which make up the piece
        """
        return self._points

    @classmethod
    @abstractmethod
    def points_from_top_left(cls, top_left: MinoPoint) -> Tuple[List[MinoPoint], Point]:
        """
        Calculates the blocks and rotational centre from the top left block coordinate
        :param top_left: the top left block coordinate
        :return: Tuple of: list of block coordinates for the piece, and the rotational centre
        """
        pass

    def rotate(self) -> None:
        """
        Rotates the piece on the board, if possible
        :return: None
        """
        new_points = [_rotate_90(p, self._centre) for p in self._points]
        if self._board.can_shift(self, new_points):
            # TODO: add "wall kick" functionality to rotate + shift the piece
            #  when its going to hit a wall or the stack from rotating
            old_points = self._points
            self._points = new_points
            # No need to update centre here
            self._board.update_piece_location(self, old_points)

    def shift(self, direction: Direction) -> bool:
        """
        Shifts the piece, if possible, in the given direction.
        :param direction: Direction
        :return: True if the piece moved successfully, False otherwise
        """
        new_points = [p.shift(direction) for p in self._points]
        if self._board.can_shift(self, new_points):
            old_points = self._points
            self._points = new_points
            self._centre = self._centre.shift(direction)
            self._board.update_piece_location(self, old_points)
            shifted = True
        else:
            shifted = False
        return shifted

    def can_shift_down(self) -> bool:
        """
        Checks if the piece can move down
        :return: True if it can shift downwards, False if not
        """
        for col in self._columns:
            lowest = self._lowest_block_in_col(col)
            if not self._board.space_below(lowest):
                return False
        return True

    @property
    def _columns(self) -> Set[int]:
        """
        The columns taken up by all the blocks that make up the piece
        :return: set of integer columns
        """
        return set(p.x for p in self._points)

    @property
    def rows(self) -> Set[int]:
        """
        The rows taken up by all the blocks that make up the piece
        :return: set of integer rows
        """
        return set(p.y for p in self._points)

    def _lowest_block_in_col(self, col: int) -> MinoPoint:
        """
        Finds the lowest block in a column, used to check ability to move down
        :param col: column value
        :return: coordinate of the lowest block in the given column for this piece
        """
        max_y = max(p.y for p in self._points if p.x == col)
        matches = [p for p in self._points if p.x == col and p.y == max_y]
        assert len(matches) == 1, f"How did we not find the lowest piece in col {col!r}?!"
        return matches[0]

    @classmethod
    @property
    @abstractmethod
    def piece_index(self) -> int:
        ...

    @classmethod
    @property
    @abstractmethod
    def letter(self) -> str:
        ...


class IPiece(Piece):
    piece_index: int = 1
    letter: str = "I"

    @classmethod
    def points_from_top_left(self, top_left: MinoPoint):
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x + 1, top_left.y),
            MinoPoint(top_left.x + 2, top_left.y),
            MinoPoint(top_left.x + 3, top_left.y),
        ]
        return points, Point(points[1].x + 0.5, top_left.y + 0.5)


class JPiece(Piece):
    piece_index: int = 2
    letter: str = "J"

    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        next_row_idx = top_left.y + 1
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x, next_row_idx),
            MinoPoint(top_left.x + 1, next_row_idx),
            MinoPoint(top_left.x + 2, next_row_idx),
        ]
        return points, points[2]


class LPiece(Piece):
    piece_index: int = 3
    letter: str = "L"

    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        next_row_idx = top_left.y + 1
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x - 2, next_row_idx),
            MinoPoint(top_left.x - 1, next_row_idx),
            MinoPoint(top_left.x, next_row_idx),
        ]
        return points, points[2]


class OPiece(Piece):
    piece_index: int = 4
    letter: str = "O"

    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x + 1, top_left.y),
            MinoPoint(top_left.x, top_left.y + 1),
            MinoPoint(top_left.x + 1, top_left.y + 1),
        ]
        return points, Point(top_left.x + 0.5, top_left.y + 0.5)



class SPiece(Piece):
    piece_index: int = 5
    letter: str = "S"

    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x + 1, top_left.y),
            MinoPoint(top_left.x - 1, top_left.y + 1),
            MinoPoint(top_left.x, top_left.y + 1),
        ]
        return points, points[3]


class TPiece(Piece):
    piece_index: int = 6
    letter: str = "T"


    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        next_row_idx = top_left.y + 1
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x - 1, next_row_idx),
            MinoPoint(top_left.x, next_row_idx),
            MinoPoint(top_left.x + 1, next_row_idx),
        ]
        return points, points[2]


class ZPiece(Piece):
    piece_index: int = 7
    letter: str = "Z"

    @classmethod
    def points_from_top_left(cls, top_left: MinoPoint):
        points = [
            MinoPoint(top_left.x, top_left.y),
            MinoPoint(top_left.x + 1, top_left.y),
            MinoPoint(top_left.x + 1, top_left.y + 1),
            MinoPoint(top_left.x + 2, top_left.y + 1),
        ]
        return points, points[2]


def _rotate_90(p: MinoPoint, centre: Point) -> MinoPoint:
    """
    Rotates a point about the origin (0,0) through 90 degrees
    :param p: the coordinate point
    :return: a new point constructed by rotating the given point
    """

    # Shift point to origin
    x = p.x - centre.x
    y = p.y - centre.y

    cos_angle = 0
    sin_angle = 1
    # Apply transformation
    new_x = x * cos_angle - y * sin_angle
    new_y = x * sin_angle + y * cos_angle

    # Shift back to original coordinates
    new_x += centre.x
    new_y += centre.y

    if new_x == p.x and new_y == p.y:
        # Don't create a whole new object for the same point
        return p
    return MinoPoint(int(new_x), int(new_y))


_SHAPE_POSSIBILITIES = [IPiece, ZPiece, LPiece, JPiece, SPiece, TPiece, OPiece]


def new_piece_type() -> Type[Piece]:
    return random.choice(_SHAPE_POSSIBILITIES)

class PieceGenerator:
    """
    Simple class used to generate and show what the next piece shape will be
    """

    def __init__(self):
        self._next_piece_type = new_piece_type()

    @property
    def next_piece_type(self) -> Type[Piece]:
        """
        The next piece type after the current
        :return: The type of piece
        """
        return self._next_piece_type

    def generate_new_piece_type(self) -> Type[Piece]:
        """
        Generates a new next piece shape
        :return: The type of piece
        """
        ret = self._next_piece_type
        self._next_piece_type = new_piece_type()
        return ret



from abc import ABC, abstractmethod
from typing import Optional
import pygame

from .board import Board
from .direction import Direction
from .command import Command
from .piece import Piece, PieceGenerator
from .interface import InterfacePygame, Interface, InterfaceCLI
from .scorer import Scorer

_LOOP_SLEEP_TIME_MS = 20


class EngineAbstract(ABC):

    @abstractmethod
    def __init__(self, board: Board, scorer: Scorer, piece_generator: PieceGenerator):
        ...

    @abstractmethod
    def run(self) -> None:
        ...


def parse_direction(key: pygame.key):
    if key == pygame.K_LEFT:
        return Direction.LEFT
    if key == pygame.K_RIGHT:
        return Direction.RIGHT
    if key == pygame.K_DOWN:
        return Direction.DOWN
    raise ValueError(f"Unsupported direction key: {key}")


class Engine(EngineAbstract):
    """
    Generalized engine class used in both versions of the game
    """
    def __init__(self, board: Board, scorer: Scorer, interface: Interface, piece_generator: PieceGenerator):
        self._board = board
        self._piece_generator = piece_generator
        self._interface = interface
        self._active_piece: Optional[Piece] = None
        self._scorer = scorer

    def run(self) -> None:
        """
        Runs the game, allowing the user to play
        :return: None
        """
        self._interface.show_instructions()
        self._run_main_loop()
        self._wait(500)
        self._run_game_over_loop()
        self._interface.quit()

    def _run_game_over_loop(self):
        """
        Shows the 'game over' screen until the player exits
        :return: None
        """
        self._interface.draw_game_over()
        show_game_over = True
        while show_game_over:
            cmds = self._interface.get_input()
            for cmd in cmds:
                if cmd == Command.QUIT:
                    show_game_over = False
                    break
            self._wait(_LOOP_SLEEP_TIME_MS)

    def _run_main_loop(self) -> None:
        """
        Runs the main gameplay logic, allowing the user to play until they lose
        :return: None
        """
        self._new_active_piece()
        self._interface.draw_screen()

        # Add timed downwards movement for passage of time
        self._set_downward_movement()

        run = True
        while run:
            need_to_refresh = False
            cmds = self._interface.get_input()
            for cmd in cmds:
                need_to_refresh = True
                if cmd == Command.HELP:
                    self._interface.show_instructions()
                    need_to_refresh = False
                    continue
                elif cmd == Command.QUIT:
                    need_to_refresh = False
                    run = False
                    break
                elif cmd == Command.PAUSE:
                    run = not self._pause()
                    need_to_refresh = run
                elif cmd == Command.ROTATE:
                    self._active_piece.rotate()
                elif cmd == Command.MOVE_BOTTOM:
                    direction = Direction.DOWN
                    while True:
                        shifted = self._active_piece.shift(direction)
                        if not shifted:
                            break
                    lines_cleared = self._board.clear_completed_rows(list(self._active_piece.rows))
                    if lines_cleared > 0:
                        levelled_up = self._scorer.add_to_score(lines_cleared)
                        if levelled_up:
                            # Refresh "falling" logic, to speed it up when needed
                            self._set_downward_movement()
                    if self._board.reached_top_row():
                        # Player has lost
                        run = False
                        break
                    self._new_active_piece()
                    break
                else:
                    direction = Direction.from_command(cmd)
                    self._active_piece.shift(direction)
                    if not self._active_piece.can_shift_down():
                        # Piece is now frozen in place
                        lines_cleared = self._board.clear_completed_rows(list(self._active_piece.rows))
                        if lines_cleared > 0:
                            levelled_up = self._scorer.add_to_score(lines_cleared)
                            if levelled_up:
                                # Refresh "falling" logic, to speed it up when needed
                                self._set_downward_movement()
                        if self._board.reached_top_row():
                            # Player has lost
                            run = False
                            break
                        self._new_active_piece()
                        break

            if need_to_refresh:
                self._interface.draw_screen()

            # Pause to share CPU - not sure if/how much it's needed, haven't
            # looked into the CPU loads yet...
            self._wait(_LOOP_SLEEP_TIME_MS)

    def _new_active_piece(self) -> None:
        """
        Updates the active piece, used when the current piece hits
        the bottom or the stack, and we need a new piece
        :return: None
        """
        self._active_piece = self._board.new_piece(self._piece_generator.generate_new_piece_type())

    @abstractmethod
    def _set_downward_movement(self) -> None:
        """
        Sets the automatic downward movement of pieces, so they fall as time passes
        :return: None
        """
        ...

    @abstractmethod
    def _wait(self, time_ms: int) -> None:
        """
        Pauses for a given number of milliseconds
        :param time_ms: time to sleep/wait in millis
        :return: None
        """
        ...

    @abstractmethod
    def _pause(self) -> bool:
        """
        Pauses the game until the pause command is seen again to unpause
        :return: True if quit command was input during pause, False otherwise
        """
        ...


class EnginePygame(Engine):
    _FALL_DELAY = 750

    def __init__(self, board: Board, scorer: Scorer, piece_generator: PieceGenerator):
        super().__init__(board, scorer, InterfacePygame(board, scorer, piece_generator), piece_generator)

    def _set_downward_movement(self) -> None:
        """
        Starts the automatic downward movement of pieces, so they fall as time passes
        :return: None
        """
        pygame.time.set_timer(pygame.event.Event(pygame.KEYDOWN, {"key": pygame.K_DOWN}), self._FALL_DELAY - self._scorer.level * 50)

    def _wait(self, time_ms: int) -> None:
        """
        Pauses for a given number of milliseconds
        :param time_ms: time to sleep/wait in millis
        :return: None
        """
        pygame.time.wait(time_ms)

    def _pause(self) -> bool:
        """
        Pauses the game until the pause command is seen again to unpause
        :return: True if quit command was input during pause, False otherwise
       """
        self._interface.show_paused()
        while True:
            cmds = self._interface.get_input()
            for cmd in cmds:
                if cmd == Command.PAUSE:
                    return False
                if cmd == Command.QUIT:
                    return True
            pygame.time.wait(50)


class EngineCLI(Engine):

    def __init__(self, board: Board, scorer: Scorer, piece_generator: PieceGenerator):
        super().__init__(board, scorer, InterfaceCLI(board, scorer, piece_generator), piece_generator)

    def _set_downward_movement(self) -> None:
        """
        Starts the automatic downward movement of pieces, so they fall as time passes
        :return: None
        """
        pass

    def _wait(self, time_ms: int) -> None:
        """
        Pauses for a given number of milliseconds
        :param time_ms: time to sleep/wait in millis
        :return: None
        """
        pass

    def _pause(self) -> bool:
        """
        Pauses the game until the pause command is seen again to unpause
        :return: True if quit command was input during pause, False otherwise
        """
        return False
